
/**************************************************************************



		This code was written by Michael Brock
			AS IS, AT YOUR OWN RISK

		Michael Brock may be contacted at :

			6524 Hagueman Road
			Richmond, VA 23225


	NOTE:  Code is property of the author.  Only god can duplicate
	it for commercial purposes. Credentials, Reason and MONEY
	required then too.

***************************************************************************/


#include <stdlib.h>
#include <alloc.h>
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <bios.h>
#include <dos.h>

#define THR 0
#define RHR 0
#define IER 1
#define IIR 2
#define LCR 3
#define MCR 4
#define LSR 5
#define MSR 6

#define RDAint 1
#define THRint 2
#define RLSint 4
#define MSint  8

#define ALLint (RDAint + THRint + RLSint + MSint)
#define THRintOFF 0xfd

#define BITS7  0x02
#define BITS8  0x03

#define STOP1  0x00
#define STOP2  0x04

#define PNONE  0x00
#define PODD   0x08
#define PEVEN  0x18

#define BAUD12 0x80
#define BAUD24 0xa0
#define BAUD48 0xc0
#define BAUD96 0xe0

#define true   1
#define false  0

typedef struct {
	unsigned char	port;
	unsigned int    base,
			mstatus,
			lstatus,

			enable_mask,
			disable_mask,
			int_number;

	char 	        *obuffer,
			*ibuffer;

	unsigned int	ohead,
			otail,
			ihead,
			itail,
			ocount,
			icount,
			omax,
			imax,

			mi,
			li,
			ri,
			ti;
}PORT;

typedef unsigned char _BOOLEAN;

PORT port, *p = &port;


void interrupt (*oldaddress)();


void interrupt far shandler(void)
{
	register int InterruptID = 0;
	register int temp;

	enable();

	for(;;)
	{
		InterruptID = inp(p->base+IIR);

		if(InterruptID & 0x0001){
			outp(0x20,0x20);
			return;
		}

		switch(InterruptID)
		{
			case 0:	p->mstatus = inp(p->base + MSR);
				p->mi = 1;
				break;
			case 2: if(p->ocount > 0)
				{
					p->ocount --;
					outp(p->base + THR,p->obuffer[p->otail]);
					p->otail++;
					if(p->otail>=p->omax) p->otail = 0;
					p->mstatus = p->lstatus = 0;
				}
				else{
					temp = inp(p->base + IER);
					if(temp & THRint) outp(p->base + IER, temp & THRintOFF);
					p->ti = 0;
				}
				break;
			case 4: temp = inp(p->base + RHR);
				if((p->imax - p->icount) > 0)
				{
					p->ibuffer[p->ohead] = (char) temp;
					if(++p->ohead >= p->imax) p->ohead = 0;
					p->icount++;
					p->mstatus = p->lstatus = 0;
					p->ri = 1;
				}else{
					//forget saving it the buffer has overrun you fool
				}
				break;
			case 6: p->lstatus = inp(p->base + LSR);
				p->li = 1;
				break;
		}
	}
}



_BOOLEAN SerialOut(int ch)
{
	register int temp;

	if((p->ocount<p->omax))
	{
		disable();
		p->obuffer[p->ohead] = ch;
		p->ocount++;
		outp(p->base + IER, ALLint); /*Turn them all on
					      just in case they have
					      been turned off*/
		enable();
		p->ti = 1;
		return (true);
	}
	p->ti = 0;
	return (false);
}

char SerialIn(void)
{
	register int temp;

	if((p->icount > 0))
	{
		disable();
		p->icount--;
		temp = p->ibuffer[p->itail];
		p->itail++;
		if(p->itail >= p->imax) p->itail = 0;
		enable();
		p->ri = 1;
		return ((char) temp);
	}
	return (0);
}

void ClearOutputBuffer (void)
{
	disable();
	p->ocount = p->ohead = p->otail = 0;
	enable();
}

void ClearInputBuffer (void)
{
	disable();
	p->icount = p->ihead = p->itail = 0;
	enable();
}

_BOOLEAN GeneralInfo(char i)
{
	return((inp(p->base + MSR) & i) ? true : false);
}

_BOOLEAN Ringing(void)
{
	return (GeneralInfo(64));
}

_BOOLEAN ClearToSend(void)
{
	return (GeneralInfo(16));
}

_BOOLEAN Carrier(void)
{
	return (GeneralInfo(32));
}

int SneakPreviewIn(void)
{
	register int temp;
	if((p->ihead != p->itail) && (p->icount > 0))
	{
		disable();
		temp = p->itail + 1;
		enable();
		if(temp > p->imax) temp = 0;
		return((char) p->ibuffer[temp]);
	}
	return (0);
}

_BOOLEAN 	SerialSetup(unsigned int pnumber, char parameters,unsigned int ob, unsigned int ib)
{
	register int imask;

	if(pnumber < 1 || pnumber > 2) return (false);

	if((p->obuffer = (char *) calloc((size_t)ob,sizeof(char)))==NULL) return (false);
	if((p->ibuffer = (char *) calloc((size_t)ib,sizeof(char)))==NULL){
		free(p->obuffer);
		return (false);
	}

	ClearOutputBuffer();
	ClearInputBuffer();
	p->port = pnumber;
	p->base = ((pnumber == 1) ? 0x3f8 : 0x2f8);
	p->omax = ob - 1;
	p->imax = ib - 1;

	p->mi = p->li = p->ri = p->ti = 0;


	switch(pnumber)
	{
		case 1:
			p->enable_mask = 0xef;
			p->disable_mask = 0x10;
			p->int_number = 4;
			break;
		case 2: p->enable_mask = 0xf7;
			p->disable_mask = 8;
			p->int_number = 3;
			break;
	}

	oldaddress = getvect((p->int_number + 8));

	disable();
	setvect((p->int_number + 8), shandler);
	enable();
	bioscom(0, parameters, pnumber-1);
	disable();
	outp(p->base + MCR, 15);
	outp(p->base + IER, ALLint);
	imask = inp(0x21) & p->enable_mask;
	outp(0x21,imask);
	enable();
	return (true);
}

_BOOLEAN EndSerial(void)
{
	register int imask;

	disable();
	outp(p->base + IER,0);
	outp(p->base + MCR,0);
	imask = inp(0x21) | p->disable_mask;
	outp(0x21,imask);
	setvect((p->int_number + 8), oldaddress);
	enable();
	free(p->obuffer);
	free(p->ibuffer);
	return (true);
}


_BOOLEAN SerialWaits(void)
{
	return((p->icount > 0) ? true : false);
}



int main(void)
{
	char ch, kb;

	if((SerialSetup(1,(BITS8 | STOP1 | PNONE | BAUD24),2048,2048)) != true) return (false);

	while(1)
	{
		kb = 0;
		if(kbhit()) kb = getch();
		if(SerialWaits()) kb = SerialIn();
		if(kb){
			putchar(kb);
			SerialOut(kb);
		}
		if(kb == 27) break;
		printf("\n|ri=%d|ti=%d|lsr=%d|msr=%d|ob=%d|ib=%d|kb=%d|ch=%d",p->ri,p->ti,p->lstatus,p->mstatus,p->ocount,p->icount,kb,ch);
	}

	EndSerial();

	return (false);
}